/*
 * wcbe_lag_frame.h
 *
 * Lag frame data types and constants, and prototypes for exported
 * functions from lag_frame.c.
 *
 * TODO: apply 'wcbe' prefix to exported symbols (and 'WCBE' to macros)
 */
#ifndef WCBE_LAG_FRAME_H_
#define WCBE_LAG_FRAME_H_

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <complex.h>
#include <sys/types.h>

G_BEGIN_DECLS

typedef u_int32_t lf_data_frame_word_t;
#define SIZEOF_LF_DATA_FRAME_WORD_T SIZEOF_U_INT32_T

typedef UINT32_T complex lf_lag_t;
# define LF_LAG_REAL(x) (__real__ (x))
# define LF_LAG_IMAG(x) (__imag__ (x))

typedef struct lf_widar_time_t {
    u_int32_t seconds;
    u_int32_t clock_cycles;
    u_int8_t epoch;
} lf_widar_time_t;

#define LF_X_INDEX 0
#define LF_Y_INDEX 1

#define LF_DATA_FRAME_MAX_LAGS 128

#define LF_BITMASK(width, offset) \
    (((((lf_data_frame_word_t)1) << (width)) - 1) << (offset))
#define LF_BITFIELD_EXTRACT(bitfield, width, offset) \
    (((bitfield) & LF_BITMASK(width, offset)) >> (offset))
#define LF_BITFIELD_PART(value, width, offset) \
    ((((lf_data_frame_word_t)(value)) & \
      ((((lf_data_frame_word_t)1) << (width)) - 1)) << (offset))

#define LF_KEY_LENGTH 4

typedef struct lf_lag_frame_t {
    lf_lag_t lags[LF_DATA_FRAME_MAX_LAGS];
    lf_data_frame_word_t key[LF_KEY_LENGTH];
    lf_widar_time_t timestamp;
    u_int32_t data_valid_center;
    u_int32_t data_valid_edge;
    u_int16_t num_lags;
    u_int8_t status;
} lf_lag_frame_t;

#define LF_LAG_FRAME_FTYPE(lf) \
    (LF_BITFIELD_EXTRACT(((lf_lag_frame_t *)(lf))->key[0], 3, 0))

typedef struct lf_cc_lag_frame_t {
    lf_lag_frame_t lf;
    int32_t data_bias;
} lf_cc_lag_frame_t;

#define LF_CC_FRAME_CCC(f) (LF_BITFIELD_EXTRACT((f)->lf.key[0], 6, 3))
#define LF_CC_FRAME_FTYPE(f) (LF_BITFIELD_EXTRACT((f)->lf.key[0], 3, 0))

#define LF_CC_FRAME_BBID_Y(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 3, 29))
#define LF_CC_FRAME_SBID_Y(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 5, 24))
#define LF_CC_FRAME_SID_Y(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 8, 16))
#define LF_CC_FRAME_BBID_X(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 3, 13))
#define LF_CC_FRAME_SBID_X(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 5, 8))
#define LF_CC_FRAME_SID_X(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 8, 0))

#define LF_CC_FRAME_LTA_BIN(f) (LF_BITFIELD_EXTRACT((f)->lf.key[2], 16, 16))
#define LF_CC_FRAME_RECIRC_BLK_Y(f) (LF_BITFIELD_EXTRACT((f)->lf.key[2], 8, 8))
#define LF_CC_FRAME_RECIRC_BLK_X(f) (LF_BITFIELD_EXTRACT((f)->lf.key[2], 8, 0))

typedef struct lf_lta_lag_frame_t {
    lf_lag_frame_t lf;
    u_int16_t frame_count;
} lf_lta_lag_frame_t;

#define LF_LTA_FRAME_START_BLOCKY(f) \
    (LF_BITFIELD_EXTRACT((f)->lf.key[0], 8, 24))
#define LF_LTA_FRAME_NBLOCKS(f) (LF_BITFIELD_EXTRACT((f)->lf.key[0], 8, 16))
#define LF_LTA_FRAME_CHIPID(f) (LF_BITFIELD_EXTRACT((f)->lf.key[0], 6, 7))
#define LF_LTA_FRAME_CCC(f) (LF_BITFIELD_EXTRACT((f)->lf.key[0], 4, 3))
#define LF_LTA_FRAME_FTYPE(f) (LF_BITFIELD_EXTRACT((f)->lf.key[0], 3, 0))

#define LF_LTA_FRAME_BBID_Y(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 3, 29))
#define LF_LTA_FRAME_SBID_Y(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 5, 24))
#define LF_LTA_FRAME_SID_Y(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 8, 16))
#define LF_LTA_FRAME_BBID_X(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 3, 13))
#define LF_LTA_FRAME_SBID_X(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 5, 8))
#define LF_LTA_FRAME_SID_X(f) (LF_BITFIELD_EXTRACT((f)->lf.key[1], 8, 0))

#define LF_LTA_FRAME_RECIRC_BLK_Y(f) \
    (LF_BITFIELD_EXTRACT((f)->lf.key[2], 8, 8))
#define LF_LTA_FRAME_RECIRC_BLK_X(f) \
    (LF_BITFIELD_EXTRACT((f)->lf.key[2], 8, 0))

#define LF_LTA_FRAME_BOARD_SN(f) (LF_BITFIELD_EXTRACT((f)->lf.key[3], 16, 16))
#define LF_LTA_FRAME_LTA_BIN(f) (LF_BITFIELD_EXTRACT((f)->lf.key[3], 16, 0))

#define LF_LTA_FRAME_LTA_BIN_BIN(f) \
    (LF_BITFIELD_EXTRACT((f)->lf.key[3], 15, 0))
#define LF_LTA_FRAME_LTA_BIN_BANK(f) \
    (LF_BITFIELD_EXTRACT((f)->lf.key[3], 1, 15))
#define LF_LTA_FRAME_PHASE_BIN(lf) \
    ((LF_LTA_FRAME_LTA_BIN_BIN(lf) \
      - LF_LTA_FRAME_RECIRC_BLK_Y(lf) + LF_LTA_FRAME_START_BLOCKY(lf)) \
     / (LF_LTA_FRAME_NBLOCKS(lf) + 1))

#define LF_DATA_FRAME_NUM_WORDS(num_lags) (11 + 2 * (num_lags))

// sync word constants are converted to network byte order
#define LF_START_SYNC_WORD (g_htonl(0xAAAAAAAA))
#define LF_END_SYNC_WORD (g_htonl(0x1C71C71C))

#define LF_CC_DATA_FRAME_FTYPE 3
#define LF_LTA_DATA_FRAME_FTYPE 4

#define LF_DATA_FRAME_FTYPE(df) (g_ntohl((df)[1]) & 0x07)

#define LF_LTA_DATA_FRAME_NUM_LAGS(df) (128 >> \
                                        ((g_ntohl((df)[1]) >> 13) & 0x07))
#define LF_LTA_DATA_FRAME_CCHIP_ID(df) ((g_ntohl((df)[1]) >> 7) & 0x3F)
#define LF_LTA_DATA_FRAME_BOARD_SN(df) ((g_ntohl((df)[8]) >> 16) & 0xFFFF)

#define LF_CC_DATA_FRAME_NUM_LAGS(df) ((g_ntohl((df)[1]) >> 9) & 0x3FF)
#define LF_CC_DATA_FRAME_CCC(df) ((g_ntohl((df)[1]) >> 3) & 0x3F)

#define LF_SUCCESS 0
#define LF_E_TOO_SMALL -100
#define LF_E_CHECKSUM -101
#define LF_E_START_SYNC -102
#define LF_E_END_SYNC -103
#define LF_E_FTYPE -104
#define LF_E_NLAGS -105
#define LF_E_STATION_ZERO -106

/* lf_lta_key_value
 *
 * Get a lag set key for the given set of (LTA) lag frame field
 * values.
 */
static inline void
lf_lta_key_value(unsigned start_block_y, unsigned num_blocks,
                 unsigned cchip_id, unsigned ccc,
                 unsigned recirculation_block_x,
                 unsigned recirculation_block_y,
                 unsigned baseband_x, unsigned baseband_y,
                 unsigned subband_x, unsigned subband_y,
                 unsigned station_x, unsigned station_y,
                 unsigned board_sn, unsigned lta_bin,
                 lf_data_frame_word_t *key)
{
    key[0] = LF_BITFIELD_PART(start_block_y, 8, 24)
        | LF_BITFIELD_PART(num_blocks - 1, 8, 16)
        | LF_BITFIELD_PART(cchip_id, 6, 7)
        | LF_BITFIELD_PART(ccc, 4, 3)
        | LF_LTA_DATA_FRAME_FTYPE;
    key[1] = LF_BITFIELD_PART(baseband_y, 3, 29)
        | LF_BITFIELD_PART(subband_y, 5, 24)
        | LF_BITFIELD_PART(station_y, 8, 16)
        | LF_BITFIELD_PART(baseband_x, 3, 13)
        | LF_BITFIELD_PART(subband_x, 5, 8)
        | LF_BITFIELD_PART(station_x, 8, 0);
    key[2] = LF_BITFIELD_PART(recirculation_block_y, 8, 8)
        | LF_BITFIELD_PART(recirculation_block_x, 8, 0);
    key[3] = LF_BITFIELD_PART(board_sn, 16, 16)
        | LF_BITFIELD_PART(lta_bin, 16, 0);
}

/* lf_cc_key_value
 *
 * Get a lag set key for the given set of correlator chip lag frame
 * field values.
 */
static inline void
lf_cc_key_value(unsigned ccc,
                unsigned baseband_x, unsigned baseband_y,
                unsigned subband_x, unsigned subband_y,
                unsigned station_x, unsigned station_y,
                unsigned lta_bin,
                unsigned recirculation_block_x, unsigned recirculation_block_y,
                lf_data_frame_word_t *key)
{
    key[0] = LF_BITFIELD_PART(ccc, 6, 3)
        | LF_CC_DATA_FRAME_FTYPE;
    key[1] = LF_BITFIELD_PART(baseband_y, 3, 29)
        | LF_BITFIELD_PART(subband_y, 5, 24)
        | LF_BITFIELD_PART(station_y, 8, 16)
        | LF_BITFIELD_PART(baseband_x, 3, 13)
        | LF_BITFIELD_PART(subband_x, 5, 8)
        | LF_BITFIELD_PART(station_x, 8, 0);
    key[2] = LF_BITFIELD_PART(lta_bin, 16, 16)
        | LF_BITFIELD_PART(recirculation_block_y, 8, 8)
        | LF_BITFIELD_PART(recirculation_block_x, 8, 0);
    key[3] = 0;
}

extern int lf_to_cc_lag_frame(lf_cc_lag_frame_t *lag_frame,
                              unsigned max_lags,
                              const lf_data_frame_word_t data_frame[]);
extern int lf_to_lta_lag_frame(lf_lta_lag_frame_t *lag_frame,
                               unsigned max_lags,
                               const lf_data_frame_word_t data_frame[]);
extern int lf_to_cc_data_frame(lf_data_frame_word_t *data_frame,
                               unsigned max_lags,
                               const lf_cc_lag_frame_t *lag_frame);
extern int lf_to_lta_data_frame(lf_data_frame_word_t *data_frame,
                                unsigned max_lags,
                                const lf_lta_lag_frame_t *lag_frame);
extern int lf_lta_lag_frame_equal(const lf_lta_lag_frame_t *frame1,
                                  const lf_lta_lag_frame_t *frame2);
extern int lf_cc_lag_frame_equal(const lf_cc_lag_frame_t *frame1,
                                 const lf_cc_lag_frame_t *frame2);
extern const char *lf_strerror(int errnum);

G_END_DECLS
#endif /* WCBE_LAG_FRAME_H_ */
